<?php

namespace Drupal\codit_batch_operations;

use Drupal\codit_batch_operations\Entity\BatchOpLog;
use Drupal\codit_batch_operations\Entity\BatchOpLogInterface;
use Drupal\Component\Render\FormattableMarkup;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Session\AccountSwitcherInterface;
use Drupal\node\NodeInterface;
use Drupal\node\NodeStorageInterface;
use Drupal\taxonomy\Entity\Term;
use Drupal\user\UserStorageInterface;
use Psr\Log\LogLevel;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class for Batch Operations.
 */
class BatchOperations implements ContainerInjectionInterface {

  /**
   * The account switcher service.
   *
   * @var \Drupal\Core\Session\AccountSwitcherInterface
   */
  protected $accountSwitcher;

  /**
   * The config factory service.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  public $entityFieldManager;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  public $entityTypeManager;

  /**
   * The name of what ran this.
   *
   * @var string
   */
  protected $executor;

  /**
   * The current BatchOpLog.
   *
   * @var \Drupal\codit_batch_operations\Entity\BatchOpLogInterface
   */
  protected $batchOpLog;

  /**
   * The user id to use during the batch operation.
   *
   * @var int
   */
  protected int $user;

  /**
   * The logger service.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  public $logger;

  /**
   * Constructs the BatchOperations service.
   *
   * @param Drupal\Core\Session\AccountSwitcherInterface $account_switcher
   *   The account switcher service.
   * @param Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory service.
   * @param Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager service.
   * @param Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   * @param Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger factory service.
   */
  public function __construct(
    AccountSwitcherInterface $account_switcher,
    ConfigFactoryInterface $config_factory,
    EntityFieldManagerInterface $entity_field_manager,
    EntityTypeManagerInterface $entity_type_manager,
    LoggerChannelFactoryInterface $logger_factory
    ) {
    $this->accountSwitcher = $account_switcher;
    $this->configFactory = $config_factory;
    $this->entityFieldManager = $entity_field_manager;
    $this->entityTypeManager = $entity_type_manager;
    $this->logger = $logger_factory->get('codit_batch_operation');
    // @todo get from config with user 1 fallback.
    $default_user = 1215;
    $this->setUser($default_user);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('account_switcher'),
      $container->get('config.factory'),
      $container->get('entity_field.manager'),
      $container->get('entity_type.manager'),
      $container->get('logger.factory')
    );
  }

  /**
   * Get the user id to use.
   *
   * @return int
   *   The user id from config or overridden.
   */
  public function getUser(): int {
    return $this->user;
  }

  /**
   * Set the user to use for batch operations.
   *
   * @param int $user
   *   The user id of the user to use.
   */
  public function setUser(int $user): void {
    $this->user = $user;
  }

  /**
   * Get the node storage.
   *
   * @return \Drupal\node\NodeStorageInterface
   *   Node storage.
   */
  public function getNodeStorage(): NodeStorageInterface {
    return $this->entityTypeManager->getStorage('node');
  }

  /**
   * Get the term storage.
   *
   * @return \Drupal\taxonomy\TermStorageInterface
   *   Term storage.
   */
  public function getTermStorage(): NodeStorageInterface {
    return $this->entityTypeManager->getStorage('taxonomy_term');
  }

  /**
   * Get the user storage.
   */
  public function getUserStorage(): UserStorageInterface {
    return $this->entityTypeManager->getStorage('user');
  }

  /**
   * Switch to the CMS Migrator user.
   *
   * @param int|null $uid
   *   The UID of the account to switch.
   */
  public function switchUser(?int $uid = NULL): void {
    // @todo connect this to config to specify the default user.
    $default_user = '';
    $uid = $uid ?? $default_user;
    $user = $this->getUserStorage()->load($uid);

    $this->accountSwitcher->switchTo($user);
    // @todo Log this to the actual log not stdout.
    debug_log_message("Acting as {$user->getDisplayName()} [{$uid}]");
  }

  /**
   * Use this for all logging.
   *
   * If call is run by CLI, it should go to stdout.  Else maybe concatted, then
   * output returned by method or dblog. But don't swamp the log.  Might need
   * two logging methods.
   */
  public function logThis(): void {
    // @todo Flesh this out.
  }

  /**
   * Load the latest revision of a node.
   *
   * @param int $nid
   *   The node ID.
   *
   * @return \Drupal\node\NodeInterface
   *   The latest revision of that node.
   */
  public function getNodeLatestRevision(int $nid): NodeInterface {
    $node_storage = get_node_storage();
    // @todo Resolve deprecation.
    return $node_storage->loadRevision($node_storage->getLatestRevisionId($nid));
  }

  /**
   * Load the default revision of a node.
   *
   * If never published, the default revision is the most recent draft.  When
   * published, the gets the most recent published revision.
   *
   * @param int $nid
   *   The node ID.
   *
   * @return \Drupal\node\NodeInterface
   *   The latest revision of that node.
   */
  public function getNodeDefaultRevision(int $nid): NodeInterface {
    return $this->getNodeStorage()->load($nid);
  }

  /**
   * Load all revisions of a node.
   *
   * @param int $nid
   *   The node ID.
   *
   * @return \Drupal\node\NodeInterface[]
   *   All revisions of that node.
   */
  public function getNodeAllRevisions(int $nid): array {
    $node_storage = $this->getNodeStorage();
    $node = $node_storage->load($nid);
    $vids = $node_storage->revisionIds($node);
    return $node_storage->loadMultipleRevisions($vids);
  }

  /**
   * Load all from the default and any forward revisons beyond that.
   *
   * @param int $nid
   *   The node ID.
   *
   * @return \Drupal\node\NodeInterface[]
   *   The default and all forward revisions of that node.
   */
  public function getNodeDefaultAndForwardRevisions(int $nid): array {
    $node_storage = $this->getNodeStorage();
    $node = $node_storage->load($nid);
    $rids = $node_storage->revisionIds($node);
    // @todo Get default revision id then use that to split the vids array.
    // $forwards = default revision id plus everthing after that
    return $node_storage->loadMultipleRevisions($forwards);
  }

  /**
   * Get an array of node ids for batch processing.
   *
   * @param string $node_bundle
   *   The bundle name of the nodes to lookup.
   * @param bool $published_only
   *   TRUE if you need only published nodes.
   *
   * @return array
   *   An array of nids for for the requested bundle, keyed by 'node_NID'.
   */
  public function getNidsOfType($node_bundle, $published_only = FALSE): array {
    $query = $this->entityTypeManager->getStorage('node')->getQuery()
      ->condition('type', $node_bundle)
      ->accessCheck(FALSE);
    if ($published_only) {
      $query->condition('status', 1);
    }

    $nids = $query->execute();
    // Having a node ids as a numeric keyed array is problematic when it comes
    // to removing things from the array. As soon as you unset one, the array
    // becomes renumbered.  So we create string keys, with numeric values.
    // [35, 75, 20] becomes
    // ['node_35' => 35, 'node_75' => 75, 'node_20' => 20].
    $node_ids = array_combine(
      array_map(self::stringify, array_values($nids)),
      array_values($nids));
    return $node_ids;
  }

  /**
   * Saves a node revision with log messaging.
   *
   * @param \Drupal\node\NodeInterface $node
   *   The node to serialize.
   * @param string $message
   *   The log message for the new revision.
   * @param bool $new
   *   Whether the revision should be created or updated.
   *
   * @return int
   *   Either SAVED_NEW or SAVED_UPDATED, depending on the operation performed.
   */
  public function saveNodeRevision(NodeInterface $node, $message = '', $new = TRUE): int {
    $moderation_state = $node->get('moderation_state')->value;
    $node->setNewRevision($new);
    $node->setSyncing(TRUE);
    $node->setValidationRequired(FALSE);
    $node->enforceIsNew(FALSE);
    // New revisions deserve special treatment.
    if ($new) {
      $node->setChangedTime(time());
      $node->setRevisionCreationTime(time());
      $uid = $this->getUser();
    }
    else {
      $uid = $node->getRevisionUserId();
      // Append new log message to previous log message.
      $prefix = !empty($message) ? $node->getRevisionLogMessage() . ' - ' : '';
      $message = $prefix . $message;
    }
    $node->setRevisionUserId($uid);
    $revision_time = $node->getRevisionCreationTime();
    // Incrementing by a nano second to bypass Drupal core logic
    // that will update the "changed" value to request time if
    // the value is not different from the original value.
    $revision_time++;
    $node->setRevisionCreationTime($revision_time);
    $node->setRevisionLogMessage($message);
    $node->set('moderation_state', $moderation_state);

    return $node->save();
  }

  /**
   * Saves a node revision with no new revision or log.
   *
   * @param \Drupal\node\NodeInterface $revision
   *   The node to serialize.
   *
   * @return int
   *   Either SAVED_NEW or SAVED_UPDATED, depending on the operation performed.
   */
  public function saveNodeExistingRevisionWithoutLog(NodeInterface $revision): int {
    $revision->setNewRevision(FALSE);
    $revision->enforceIsNew(FALSE);
    $revision->setSyncing(TRUE);
    $revision->setValidationRequired(FALSE);
    $revision_time = $revision->getRevisionCreationTime();
    // Incrementing by a nano second to bypass Drupal core logic
    // that will update the "changed" value to request time if
    // the value is not different from the original value.
    $revision_time++;
    $revision->setRevisionCreationTime($revision_time);
    $revision->setChangedTime($revision_time);
    return $revision->save();
  }

  /**
   * Create new terms for if they do not exist.
   *
   * @param string $vocabulary_id
   *   The machine name of the taxonomy vocabulary.
   * @param array $terms
   *   An array of terms in the form of 'term name' => 'description'.
   *
   * @return int
   *   The number of terms created.
   */
  public function saveNewTerms($vocabulary_id, array $terms): int {
    $terms_created = 0;
    foreach ($terms as $name => $description) {
      // Make sure we are not creating duplicate terms.
      $tid = $this->entityTypeManager()->getStorage('taxonomy_term')->getQuery()
        ->condition('name', $name)
        ->condition('vid', $vocabulary_id)
        ->accessCheck(FALSE)
        ->execute();
      if (empty($tid)) {
        // Term does not exist, so create it.
        $term = Term::create([
          'name' => $name,
          'vid' => $vocabulary_id,
        ]);
        $term->setNewRevision(TRUE);
        $term->setDescription($description);
        $term->setRevisionUserId($this->getUser());
        $term->setSyncing(TRUE);
        $term->setValidationRequired(FALSE);
        $term->save();
        $terms_created++;
      }
    }
    return $terms_created;
  }

  /**
   * Initializes the basic sandbox values.
   *
   * @param array $sandbox
   *   Standard drupal $sandbox var to keep state in hook_update_N.
   *
   * @throws Drupal\Core\Utility\UpdateException
   *   If the counter callback can not be found.
   */
  public function initSandbox(array &$sandbox) {
    if (empty($sandbox['total'])) {
      // Sandbox has not been initiated.
      $sandbox['items_to_process'] = $this->getItemsToProcess();
      $sandbox['total'] = count($sandbox['items_to_process']);
      $sandbox['current'] = 0;
      $sandbox['BatchOpLog'] = $this->getBatchOpLog(TRUE);
      $script_name = get_class($this);
      $sandbox['multi_run_state_key'] = "cbo_{$script_name}";
      // This seems like the first run, see if there is already a state saved
      // from a previous attempt.
      $last_run_completed = \Drupal::state()->get($sandbox['multi_run_state_key']);
      if (!is_null($last_run_completed)) {
        // A state exists, so alter the 'current' and 'items_to_process'.
        $sandbox['current'] = $last_run_completed + 2;
        // Remove the last successful run, and all that came before it.
        $sandbox['items_to_process'] = array_slice($sandbox['items_to_process'], $last_run_completed);
      }
    }
    $this->batchOpLog = $sandbox['BatchOpLog'];
    $sandbox['element'] = array_key_first($sandbox['items_to_process']);
  }

  /**
   * Updates the counts and log if complete.
   *
   * @param array $sandbox
   *   Hook_update_n sandbox for keeping state.
   *
   * @return string
   *   String to be used as update hook messages.
   */
  public function completeSandbox(array &$sandbox) {
    // Determine when to stop batching.
    $sandbox['current'] = ($sandbox['total'] - count($sandbox['items_to_process']));
    // Save the 'current' value to state, to record a successful run.
    \Drupal::state()->set($sandbox['multi_run_state_key'], $sandbox['current']);
    $sandbox['#finished'] = (empty($sandbox['total'])) ? 1 : ($sandbox['current'] / $sandbox['total']);
    $vars = [
      '@completed' => $sandbox['current'],
      '@element' => $sandbox['element'],
      '@total' => $sandbox['total'],
    ];

    $message = t('Processed @element. @completed/@total.', $vars) . PHP_EOL;
    // Log the all finished notice.
    if ($sandbox['#finished'] === 1) {
      $this->logger->log(LogLevel::INFO, $this->getCompletedMessage(), $vars);
      $logged_message = new FormattableMarkup($this->getCompletedMessage(), $vars);
      $message = t('Process completed:') . " {$logged_message}" . PHP_EOL;
      // Delete the state as it is no longer needed.
      \Drupal::state()->delete($sandbox['multi_run_state_key']);
      $this->batchOpLog->setCompleted()
        ->appendLog($message)
        ->save();
    }
    return $message;
  }

  /**
   * Lookup a key in a map array and return the value from the map.
   *
   * @param string|null $lookup
   *   A map key to lookup. Do not lookup int as indexes can shift.
   * @param array $map
   *   An array containing string key value pairs. [lookup => value].
   * @param bool $strict
   *   TRUE = only want a value from the array, FALSE = want your lookup back.
   *
   * @return mixed
   *   Whatever the value associated with the key.
   */
  public function mapToValue(string|null $lookup, array $map, bool $strict = TRUE) : mixed {
    if (empty($lookup)) {
      if (isset($map['default'])) {
        // There is a default set, so use it.
        return $map['default'];
      }
      elseif ($strict) {
        return NULL;
      }
      else {
        return $lookup;
      }
    }
    if ($strict) {
      // Strict, so either it is there, or nothing.
      return $map[$lookup] ?? NULL;
    }
    else {
      // Not strict, so pass back what given it its not in the map.
      return $map[$lookup] ?? $lookup;
    }
  }

  /**
   * Modifies to items to process to make sure we have string keys.
   *
   * @return array
   *   The items to be processed.
   */
  protected function getItemsToProcess(): array {
    $items = $this->gatherItemsToProcess();
    // Having flat arrays or numeric keyed array is problematic when it comes
    // to removing things from the array. As soon as you unset one, the array
    // becomes renumbered.  So we create string keys, with numeric values.
    // [35, 75, 20] becomes
    // ['item_35' => 35, 'item_75' => 75, 'item_20' => 20].
    $properly_keyed = array_combine(
      array_map(['Drupal\codit_batch_operations\BatchOperations', 'stringify'], array_keys($items)),
      array_values($items));
    return $properly_keyed;
  }

  /**
   * Callback function to convert an int to a string for keys.
   *
   * @param int|string $id
   *   A numeric id like a nid, tid or pid.
   *
   * @return string
   *   The id concatenated to the end of item_.
   */
  public static function stringify($id) {
    return "item_{$id}";
  }

  /**
   * Method to run the whole batch operation.
   *
   * @param array $sandbox
   *   The snadbox by reference used to keep state across runs.
   * @param string $executor
   *   The thing that is executing the scripts (hook_update, deploy).
   *
   * @return string
   *   Status and closing messages for each run.
   */
  public function run(array &$sandbox, string $executor):string {
    $this->executor = $executor;
    $this->initSandbox($sandbox);
    $batch_items = array_slice($sandbox['items_to_process'], 0, $this->getBatchSize(), TRUE);
    if ($this->getBatchSize() > 1) {
      // We are using batches, so let's separate them in the log.
      $this->batchOpLog->appendLog('----- batch -----');
    }
    foreach ($batch_items as $key => $item) {
      try {
        $message = $this->processOne($key, $item, $sandbox);
        $this->batchOpLog->appendLog("{$key}: {$message}");
        unset($sandbox['items_to_process'][$key]);
      }
      catch (\Throwable $th) {
        $this->batchOpLog->appendError($th->getMessage())
          ->save();
        throw new \Exception($th->getMessage());
      }
    }
    $this->batchOpLog->save();

    return $this->completeSandbox($sandbox);
  }

  /**
   * Gets the BatchOpLog and sets it if it does not exist.
   *
   * @param bool $new
   *   A flag to force creating a new BatchOpLog.
   *
   * @return \Drupal\codit_batch_operations\Entity\BatchOpLogInterface
   *   The BatchOpLog.
   */
  public function getBatchOpLog(bool $new = FALSE): BatchOpLogInterface {
    if (empty($this->batchOpLog) || $new) {
      // It doesn't exist so instantiate it.
      $this->batchOpLog = BatchOpLog::create([
        'name' => get_class($this),
        'user_id' => $this->getUser(),
        'executor' => $this->executor,
        'created' => time(),
        'completed' => FALSE,
      ]);
    }
    return $this->batchOpLog;
  }

}
